# Imports
import meep as mp 
#from meep.materials import SiO2, Si
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from matplotlib.animation import FFMpegWriter
import logging
import sys

# Redirect stdout and stderr to logging
class LoggerWriter:
    def __init__(self, level):
        self.level = level
        self.buffer = ""  # Buffer to handle partial lines

    def write(self, message):
        # Accumulate messages until a full line (ending in \n) is received
        self.buffer += message
        while "\n" in self.buffer:
            line, self.buffer = self.buffer.split("\n", 1)
            if line.strip():  # Only log non-empty lines
                self.level(line)

    def flush(self):
        # Log any remaining messages in the buffer
        if self.buffer.strip():
            self.level(self.buffer.strip())
        self.buffer = ""  # Clear the buffer


def output_to_file(rel_path = "output.log"):
    logging.basicConfig(
        filename=rel_path,  # Log file
        level=logging.INFO,  # Logging level
        format='%(asctime)s - %(levelname)s - %(message)s',  # Log format with timestamp
        datefmt='%Y-%m-%d %H:%M:%S'  # Date and time format
    )

    # Redirect `print` and errors
    sys.stdout = LoggerWriter(logging.info)  # Redirect standard output to logging
    sys.stderr = LoggerWriter(logging.error)  # Redirect errors to logging


class SimulationHandler:
    def __init__(self):
        self.Ex_list = []
        self.Ey_list = []
        self.Ez_list = []
        self.Hx_list = []
        self.Hy_list = []
        self.Hz_list = []

    def update_fields(self, ex, ey, ez, hx, hy, hz):
        self.Ex_list.append(ex[0])
        self.Ey_list.append(ey[0])
        self.Ez_list.append(ez[0])
        self.Hx_list.append(hx[0])
        self.Hy_list.append(hy[0])
        self.Hz_list.append(hz[0])
    
    def save_data(self, cell_x, cell_y):
        Ex = np.array(self.Ex_list)
        Ey= np.array(self.Ey_list)
        Ez = np.array(self.Ez_list)
        Hx = np.array(self.Hx_list)
        Hy= np.array(self.Hy_list)
        Hz = np.array(self.Hz_list)
        np.savez(f'updated/fieldData.npz', Ex=Ex, Ey=Ey, Ez=Ez, Hx=Hx, Hy=Hy, Hz=Hz, cell_x=cell_x, cell_y=cell_y)
    
    def visualize_field(self,cell_x , cell_y, component="Ex", show=False):
        if component == "Ex":
            data = np.array(self.Ex_list)
        elif component == "Ey":
            data = np.array(self.Ey_list)
        elif component == "Ez":
            data = np.array(self.Ez_list)
        elif component == "Hx":
            data = np.array(self.Hx_list)
        elif component == "Hy":
            data = np.array(self.Hy_list)
        elif component == "Hz":
            data = np.array(self.Hz_list)
        
        def animate(i):
            im.set_data(data[i].T)
            return im,

        max = np.max(data)
        min = np.min(data)

        fig, ax = plt.subplots()
        im = ax.imshow(data[0].T, cmap='RdBu', vmin=min, vmax=max, extent=[cell_x/2, -cell_x/2, cell_y/2, -cell_y/2])
        fig.colorbar(im)
        numberAnimationTimesteps = len(data)
        fieldAnimation = animation.FuncAnimation(fig, animate, frames=numberAnimationTimesteps, interval=20, blit=True)

        fieldAnimation.save(f'coupler_{component}.gif', writer='imagemagick')

        writer = FFMpegWriter(fps=30, codec="libx264")

        fieldAnimation.save(f'coupler_{component}.mp4', writer=writer)
        if show:
            plt.show()

    
    def visualize_fields(self, show=False):
        self.visualize_field(component="Ex", show=show)
        self.visualize_field(component="Ey", show=show)
        self.visualize_field(component="Ez", show=show)
        self.visualize_field(component="Hx", show=show)
        self.visualize_field(component="Hy", show=show)
        self.visualize_field(component="Hz", show=show)
        

output_to_file("updated/coupler_output.log")

res = 50  # pixels/μm
is_3D = False  # 3d calculation?
gdsII_file = "hybrid_coupler.gds"

# Materials used in the simulation
Si = mp.Medium(index=3.45)
SiO2 = mp.Medium(index=1.444)

#dimensions of the coupler that is to be simulated
coupler_length = 28
coupler_width = 4.5
port_separation = 4
si_thickness = 0.22

#dimensions of the simulation
sim_thickness = 2
pml_thickness = 2

# Set up frequency points for simulation
npoints = 20
lcen = 1.55  #central wavelength
dwl = 0.05
fcen = 1 / lcen
df = 1 / dwl
df_source =  3 * dwl / lcen**2  # ONLY USE FOR SOURCES TO ENSURE LARGE ENOUGH FREQUENCY BANDWIDTH FOR SIMULATION
wl = np.linspace(lcen - dwl/2, lcen + dwl/2, npoints)
fpoints = 1 / wl


cell_thickness = pml_thickness + sim_thickness + pml_thickness if is_3D else 0
si_thickness = si_thickness if is_3D else 20

# Sets the min and max values for the cell and the silicon. Our simulation will be centered at y=0
si_zmax = 0.5 * si_thickness
si_zmin = -0.5 * si_thickness

#import geometry from gds file
geometry = mp.get_GDSII_prisms(Si, gdsII_file, 1, si_zmin, si_zmax)

#define simulation volume
cell = mp.Vector3(coupler_length + 2*pml_thickness + 1,coupler_width+2*pml_thickness + 2,cell_thickness)
cell_center = mp.Vector3(0,0,0)

#define location and geometry of source and monitors
port_size = mp.Vector3(0,1,1)
port_dx = coupler_length/2
port_dy = port_separation/2
port1 = mp.Volume(center=mp.Vector3(-port_dx,-port_dy,0), size=port_size)
port2 = mp.Volume(center=mp.Vector3(-port_dx,port_dy,0), size=port_size)
port3 = mp.Volume(center=mp.Vector3(port_dx,port_dy,0), size=port_size)
port4 = mp.Volume(center=mp.Vector3(port_dx,-port_dy,0), size=port_size)
source = mp.Volume(center=mp.Vector3(-port_dx-0.25,-port_dy,0),size=port_size)

mode_parity = mp.NO_PARITY if is_3D else mp.EVEN_Y + mp.ODD_Z

# setup source
sources = [
    mp.EigenModeSource(
        src=mp.GaussianSource(wavelength=lcen, fwidth=df_source),
        size=source.size,
        center=source.center,
        eig_band=1,
        eig_parity=mode_parity,
        eig_match_freq=True,
    )
]

#setup simulation
sim = mp.Simulation(
    resolution=res,
    cell_size=cell,
    boundary_layers=[mp.Absorber(pml_thickness)],
    sources=sources,
    geometry=geometry,
    default_material=SiO2
)


# Adds mode monitors at each of the ports to track the energy that goes in or out
mode_monitor_1 = sim.add_mode_monitor(fpoints, mp.ModeRegion(volume=port1))
mode_monitor_2 = sim.add_mode_monitor(fpoints, mp.ModeRegion(volume=port2))
mode_monitor_3 = sim.add_mode_monitor(fpoints, mp.ModeRegion(volume=port3))
mode_monitor_4 = sim.add_mode_monitor(fpoints, mp.ModeRegion(volume=port4))

plot_plane = mp.Volume(center=cell_center, size=mp.Vector3(cell.x, cell.y, 0))
sim.plot2D(output_plane=plot_plane if is_3D else None) # No parameters are needed for a 2D simulation. 
#plt.savefig("updated/sim_setup.png")
plt.show()


sim_handler = SimulationHandler()

def updateField(sim):
    ex = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Ex)])
    ey = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Ey)])
    ez = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Ez)])
    hx = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Hx)])
    hy = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Hy)])
    hz = np.array([sim.get_array(center=cell_center,size=mp.Vector3(cell.x, cell.y, 0),component=mp.Hz)])
    sim_handler.update_fields(ex, ey, ez, hx, hy, hz)

sim.run(mp.at_every(1, updateField), until=260)


sim_handler.save_data(cell.x, cell.y)

# Finds the S parameters
norm_mode_coeff = sim.get_eigenmode_coefficients(mode_monitor_1, [1], eig_parity=mode_parity).alpha[0, :, 0]#
port1_coeff = sim.get_eigenmode_coefficients(mode_monitor_1, [1], eig_parity=mode_parity).alpha[0, :, 1] / norm_mode_coeff
port2_coeff = sim.get_eigenmode_coefficients(mode_monitor_2, [1], eig_parity=mode_parity).alpha[0, :, 1] / norm_mode_coeff
port3_coeff = sim.get_eigenmode_coefficients(mode_monitor_3, [1], eig_parity=mode_parity).alpha[0, :, 0] / norm_mode_coeff
port4_coeff = sim.get_eigenmode_coefficients(mode_monitor_4, [1], eig_parity=mode_parity).alpha[0, :, 0] / norm_mode_coeff

np.savez('updated/s_params.npz', wl=wl, s11=port1_coeff, s21=port2_coeff, s31=port3_coeff, s41=port4_coeff)

print("#######################################################################################################")

